/*******************************************************************
 *	Copyright (c) 1994-1998 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File:		driver/bc_alg.c
 *
 *	Description:	algorithm wrapper for module building
 *
 *	Scope:		BestCrypt algorithms drivers
 *
 *	Platforms:	Linux
 * 
 *	Author:		Nail R. Kaipov
 *
 *	Created:	10-Nov-1998
 *
 *	Revision:	
 *
 *******************************************************************/

#include "../bc_cfg.h"
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <asm/uaccess.h>
#include "bc_types.h"
#include "../bc_mgr.h"		/* algorithm manager defines	*/
#include "algorithm.h"		/* algorithm defines		*/

#define POOL_SIZE_BYTES		512
#define TABLE_LENGTH 32

typedef struct {
	unsigned char	*key_extended;
	unsigned char	*pool;
	int		lock;
} KEY_ELEMENT;

static KEY_ELEMENT KeyTable[ TABLE_LENGTH ];

/*-- keys management -------------------------------------------*/
static void shred (u_char **ptr, u_int len)
{
	if (!ptr || !*ptr) return;
	memset(*ptr, 0, len);
	kfree(*ptr);
	*ptr = NULL;
}

int test_key (KEY_HANDLE kh)
{ 
	if (kh < 0 || kh >= TABLE_LENGTH) return -EINVAL;
        if (!KeyTable[kh].key_extended)   return -EINVAL;
	return 0;
}

int lock_key (KEY_HANDLE kh, int lock)
{ 
	int error;
	
	error = test_key(kh);
	if (error) return error;
	KeyTable[kh].lock = lock;
	return 0;
}

int free_key (KEY_HANDLE kh) 
{
	int err;

	err = test_key(kh);
	if (err) return err;
	if (KeyTable[kh].lock) return 0;
	
	shred(&(KeyTable[kh].key_extended), KEY_LENGTH_EXTENDED/8);
	shred(&(KeyTable[kh].pool), POOL_SIZE_BYTES);

	memset(&(KeyTable[kh]), 0, sizeof(KEY_ELEMENT));

	MOD_DEC_USE_COUNT;
	return 0;
}

int make_key(u_char *key, u_int keylen, u_char *pool, KEY_HANDLE *return_key)
/* NOTE!!! key and pool are in user memory space. */
{
	KEY_HANDLE	kh = -1;
	int		error = 0; //, i;
	u_char		*key_tmp = NULL;

	if (keylen != KEY_LENGTH) return -EINVAL;
	keylen = (keylen+7) >> 3; /* convert to bytes */
	for (kh = 0; kh < TABLE_LENGTH; kh++)
		if (!KeyTable[kh].key_extended) break;

	if (kh >= TABLE_LENGTH || KeyTable[kh].key_extended || KeyTable[kh].pool)
		return -EBUSY;

	error = -ENOMEM;

	KeyTable[kh].key_extended = kmalloc(KEY_LENGTH_EXTENDED/8, GFP_KERNEL);
	key_tmp = kmalloc(keylen, GFP_KERNEL);
	KeyTable[kh].pool = kmalloc(POOL_SIZE_BYTES, GFP_KERNEL);

	if (!KeyTable[kh].key_extended || !key_tmp || !KeyTable[kh].pool) 
		goto error_out;

	error = copy_from_user(KeyTable[kh].pool, pool, POOL_SIZE_BYTES);
	if (error) 
		goto error_out;
	
	error = copy_from_user(key_tmp, key, keylen);
	if (error) 
		goto error_out;
	
	KeyExtend(key_tmp, (DWORD *)KeyTable[kh].key_extended);
	shred(&key_tmp, keylen);

	*return_key = kh;
	MOD_INC_USE_COUNT;	
	return 0;
	
error_out:
	shred(&(KeyTable[kh].key_extended), KEY_LENGTH_EXTENDED/8);
	shred(&(KeyTable[kh].pool), POOL_SIZE_BYTES);
	shred(&key_tmp, keylen);
	return error;		
}


/*-- encryption/decryption helpers -----------------------------*/

inline void transform_iv (DWORD *iv, DWORD *pool)
{
	unsigned int i;
	
	i = (iv[0] & 0x3F) << 1;
	iv[0] ^= pool[i];
	iv[1] ^= pool[i+1];
}

inline void transform_buf (DWORD *pool, DWORD *src, DWORD *dst, DWORD len)
{
	while (len--) *dst++ = *src++ ^ *pool++;
}

/*-- encryption/decryption main --------------------------------*/

int encrypt (KEY_HANDLE kh, u_char *iv, u_char *src, u_char *dst, u_int len)
{
	int	err;
	WORD	*s, ivector[4], ivectorS[4];
	DWORD	i, offset, size, *pool;

	if (!len) return 0;
	if (len & BLOCK_LENGTH) return -EINVAL;
	err = test_key(kh);
	if (err) return err;
	
	pool = (DWORD *)KeyTable[kh].pool;
	s = (WORD *)iv;
	for (i = 0; i < 4; i++) ivector[i] = ivectorS[i] = s[i];
	
	for (offset = 0; ; offset += 512) {
		transform_iv((DWORD *)ivector, pool);
		size = len > 512 ? 512 : len;
		len -= size;
		transform_buf(pool, 
			(DWORD *)(&src[offset]), 
			(DWORD *)(&dst[offset]), 
			size >> 2);
		Encrypt((DWORD *)ivector,
			(DWORD *)KeyTable[kh].key_extended,
			(DWORD *)(&dst[offset]),
			(DWORD *)(&dst[offset]),
			size);
		if (!len) break;
		/* We need to update IVector every 512 bytes */
		for(i = 0; i < 4; i++) ivector[i] = ++(ivectorS[i]); 
	}
	return 0;
}

int decrypt (KEY_HANDLE kh, u_char *iv, u_char *src, u_char *dst, u_int len)
{
	int	err;
	WORD	*s, ivector[4], ivectorS[4];
	DWORD	i, offset, size, *pool;
			
	if (!len) return 0;
	err = test_key(kh);
	if (err) return err;
	
	pool = (DWORD *)KeyTable[kh].pool;
	s = (WORD *)iv;
	for (i = 0; i < 4; i++) ivector[i] = ivectorS[i] = s[i];
	
	for (offset = 0; ; offset += 512) {
		transform_iv((DWORD *)ivector, pool);
		size = len > 512 ? 512 : len;
		len -= size;
		Decrypt((DWORD *)ivector,
			(DWORD *)KeyTable[kh].key_extended,
			(DWORD *)(&src[offset]),
			(DWORD *)(&dst[offset]),
			size);
		transform_buf(pool, 
			(DWORD *)(&dst[offset]), 
			(DWORD *)(&dst[offset]), 
			size >> 2);
		if (!len) break;
		/* We need to update IVector every 512 bytes */
		for(i = 0; i < 4; i++) ivector[i] = ++(ivectorS[i]); 
	}
	return 0;
}

/*-- modularization --------------------------------------------*/

EXPORT_NO_SYMBOLS;

static struct bc_algorithm algo;

int init_module (void)
{
	int i;
	
	for (i = 0; i < TABLE_LENGTH; i++)
		memset(&KeyTable[i], 0x00, sizeof(KEY_ELEMENT));
	
	algo.id       = ALGORITHM_ID;	
	algo.name     = DISPLAY_NAME;
	algo.keylen   = KEY_LENGTH;
	algo.make_key = make_key;
	algo.test_key = test_key;
	algo.free_key = free_key;
	algo.lock_key = lock_key;
	algo.encrypt  = encrypt;
	algo.decrypt  = decrypt;
	
	i = register_bc_algo(&algo);	
	return i;
}

void cleanup_module (void) 
{
	int i;
	
	for (i = 0; i < TABLE_LENGTH; i++) {
		if (KeyTable[i].key_extended) {
			memset(KeyTable[i].key_extended, 0, KEY_LENGTH_EXTENDED/8);
			kfree (KeyTable[i].key_extended);
		}
		if (KeyTable[i].pool) {
			memset(KeyTable[i].pool, 0, POOL_SIZE_BYTES);
			kfree (KeyTable[i].pool);
		}
		memset(&KeyTable[i], 0, sizeof(KEY_ELEMENT));
	}
	i = unregister_bc_algo(ALGORITHM_ID);
	return;
}
